/*
 * Segment 6 Intelligente Bewertungsverfahren, SWE 6.2 Baustellensimulation
 * Copyright (C) 2007 BitCtrl Systems GmbH
 *
 * This program is free software; you can redistribute it and/or modify it under
 * the terms of the GNU General Public License as published by the Free Software
 * Foundation; either version 2 of the License, or (at your option) any later
 * version.
 *
 * This program is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
 * FOR A PARTICULAR PURPOSE. See the GNU General Public License for more
 * details.
 *
 * You should have received a copy of the GNU General Public License along with
 * this program; if not, write to the Free Software Foundation, Inc., 51
 * Franklin Street, Fifth Floor, Boston, MA 02110-1301, USA.
 *
 * Contact Information:
 * BitCtrl Systems GmbH
 * Weißenfelser Straße 67
 * 04229 Leipzig
 * Phone: +49 341-490670
 * mailto: info@bitctrl.de
 */

package de.bsvrz.ibv.simbaustelle.parameter;

import javax.swing.event.EventListenerList;

import com.bitctrl.Constants;

import de.bsvrz.dav.daf.main.ClientDavInterface;
import de.bsvrz.dav.daf.main.ClientReceiverInterface;
import de.bsvrz.dav.daf.main.Data;
import de.bsvrz.dav.daf.main.DataDescription;
import de.bsvrz.dav.daf.main.ReceiveOptions;
import de.bsvrz.dav.daf.main.ReceiverRole;
import de.bsvrz.dav.daf.main.ResultData;
import de.bsvrz.dav.daf.main.config.Aspect;
import de.bsvrz.dav.daf.main.config.AttributeGroup;
import de.bsvrz.dav.daf.main.config.ConfigurationObject;
import de.bsvrz.ibv.simbaustelle.simbaustelle.BssKonstanten;
import de.bsvrz.sys.funclib.debug.Debug;

/**
 * Die Repräsenation der Parameterverwaltung für die Baustellensimulation eines
 * VerkehrsmodellNetzes.
 *
 * @author BitCtrl Systems GmbH, Peuker
 */
public class BssParameterObserver implements IBssParameterObserver, ClientReceiverInterface {

	/**
	 * der Logger für Debug-Ausgaben.
	 */
	private static final Debug LOGGER = Debug.getLogger();
	/**
	 * der aktuelle Parameterdatensatz.
	 */
	private Data daten;
	/**
	 * die Liste der angemeldeten Listener, die bei der Änderung des
	 * Parameterdatensatzes benachrichtigt werden sollen.
	 */
	private final EventListenerList listenerListe = new EventListenerList();

	/**
	 * erzeugt eine Instanz der Parameterverwaltung.
	 *
	 * @param connection
	 *            die verwendete Datenverteilerverbindung
	 * @param netz
	 *            das zu verwaltende VerkehrsmodellNetz
	 */
	public BssParameterObserver(final ClientDavInterface connection, final ConfigurationObject netz) {
		final AttributeGroup atg = connection.getDataModel()
				.getAttributeGroup(BssKonstanten.PID_ATG_BAUSTELLEN_SIMULATION_MODELL);
		final Aspect asp = connection.getDataModel().getAspect("asp.parameterSoll");
		final DataDescription dataDescription = new DataDescription(atg, asp);

		final ResultData parameter = connection.getData(netz, dataDescription, 0);
		if ((parameter == null) || (parameter.getData() == null)) {
			throw new IllegalStateException(
					"Es ist kein Parameter für die Baustellensimulation definiert, Objekt: " + netz);
		}
		daten = parameter.getData().createUnmodifiableCopy();
		connection.subscribeReceiver(this, netz, dataDescription, ReceiveOptions.normal(), ReceiverRole.receiver());
	}

	@Override
	public final void addParameterListener(final IBssParameterListener listener) {
		synchronized (listenerListe) {
			listenerListe.add(IBssParameterListener.class, listener);
		}
	}

	@Override
	public final long getAktualisierungsIntervall() {
		return daten.getUnscaledValue(BssKonstanten.PARAM_NAME_AKTUALISIERUNGS_INTERVALL).longValue()
				* Constants.MILLIS_PER_DAY;
	}

	@Override
	public final double getFaktorQ0() {
		return daten.getScaledValue(BssKonstanten.PARAM_NAME_FAKTOR_Q0).doubleValue();
	}

	@Override
	public final long getHorizont() {
		return daten.getUnscaledValue(BssKonstanten.PARAM_NAME_PROGNOSE_HORIZONT).longValue()
				* Constants.MILLIS_PER_DAY;
	}

	@Override
	public final long getLaengePkw() {
		return daten.getUnscaledValue(BssKonstanten.PARAM_NAME_LAENGE_PKW).longValue();
	}

	@Override
	public final long getLaengeStauAufhebung() {
		return daten.getUnscaledValue(BssKonstanten.PARAM_NAME_LAENGE_STAU_AUFHEBUNG).longValue();
	}

	@Override
	public final long getLaengeStauBeginn() {
		return daten.getUnscaledValue(BssKonstanten.PARAM_NAME_LAENGE_STAU_BEGINN).longValue();
	}

	/**
	 * gibt den Parameterdatensatz als Zeichenkette aus.
	 *
	 * @return den Text
	 */
	private String getParameterDescription() {
		final StringBuffer buffer = new StringBuffer();

		buffer.append(Constants.NL);
		buffer.append("PrognoseHorizont:");
		buffer.append(getHorizont() / Constants.MILLIS_PER_DAY);
		buffer.append(" Tage");
		buffer.append(Constants.NL);

		buffer.append("AktualisierungsIntervall:");
		buffer.append(getAktualisierungsIntervall() / Constants.MILLIS_PER_DAY);
		buffer.append(" Tage");
		buffer.append(Constants.NL);

		buffer.append("IterationsSchrittweite:");
		buffer.append(getPrognoseSchrittweite());
		buffer.append(" ms");
		buffer.append(Constants.NL);

		buffer.append("FaktorQ0:");
		buffer.append(getFaktorQ0());
		buffer.append(" %");
		buffer.append(Constants.NL);

		buffer.append("LängePkw:");
		buffer.append(getLaengePkw());
		buffer.append(" m");
		buffer.append(Constants.NL);

		buffer.append("LängeStauBeginn:");
		buffer.append(getLaengeStauBeginn());
		buffer.append(" m");
		buffer.append(Constants.NL);

		buffer.append("LängeStauAufhebung:");
		buffer.append(getLaengeStauAufhebung());
		buffer.append(" m");

		return buffer.toString();
	}

	@Override
	public final long getPrognoseSchrittweite() {
		return daten.getTimeValue(BssKonstanten.PARAM_NAME_ITERATIONS_SCHRITTWEITE).getMillis();
	}

	@Override
	public final void removeParameterListener(final IBssParameterListener listener) {
		synchronized (listenerListe) {
			listenerListe.add(IBssParameterListener.class, listener);
		}
	}

	@Override
	public void update(final ResultData[] results) {
		for (final ResultData result : results) {
			this.daten = null;
			if (result.getData() != null) {
				this.daten = result.getData().createUnmodifiableCopy();
				LOGGER.finer("Neue Parameter empfangen: " + getParameterDescription());
				synchronized (listenerListe) {
					for (final IBssParameterListener listener : listenerListe
							.getListeners(IBssParameterListener.class)) {
						listener.changed(this);
					}
				}
			}
		}
	}
}
