/*
 * Segment 6 Intelligente Bewertungsverfahren, SWE 6.2 Baustellensimulation
 * Copyright (C) 2007 BitCtrl Systems GmbH
 *
 * This program is free software; you can redistribute it and/or modify it under
 * the terms of the GNU General Public License as published by the Free Software
 * Foundation; either version 2 of the License, or (at your option) any later
 * version.
 *
 * This program is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
 * FOR A PARTICULAR PURPOSE. See the GNU General Public License for more
 * details.
 *
 * You should have received a copy of the GNU General Public License along with
 * this program; if not, write to the Free Software Foundation, Inc., 51
 * Franklin Street, Fifth Floor, Boston, MA 02110-1301, USA.
 *
 * Contact Information:
 * BitCtrl Systems GmbH
 * Weißenfelser Straße 67
 * 04229 Leipzig
 * Phone: +49 341-490670
 * mailto: info@bitctrl.de
 */

package de.bsvrz.ibv.simbaustelle.prognose.bss;

import java.text.MessageFormat;

import de.bsvrz.dav.daf.main.ClientDavInterface;
import de.bsvrz.ibv.simbaustelle.parameter.IBssParameterObserver;
import de.bsvrz.ibv.simbaustelle.prognose.BaustellenSimulator;
import de.bsvrz.ibv.simbaustelle.prognose.BerechnungsZeitRaum;
import de.bsvrz.ibv.simbaustelle.prognose.BerechnungsZeitRaumSchnittstelle;
import de.bsvrz.ibv.simbaustelle.prognose.ZufliessenderVerkehr;
import de.bsvrz.ibv.simbaustelle.prognose.ZufliessenderVerkehrSchnittstelle;
import de.bsvrz.ibv.simbaustelle.simbaustelle.BssMessage;
import de.bsvrz.ibv.simbaustelle.verwaltung.BssStau;
import de.bsvrz.ibv.simbaustelle.verwaltung.SimulationsAuftrag;
import de.bsvrz.sys.funclib.bitctrl.daf.BetriebsmeldungDaten;
import de.bsvrz.sys.funclib.bitctrl.daf.LogTools;
import de.bsvrz.sys.funclib.bitctrl.modell.verkehr.objekte.MessQuerschnittAllgemein;
import de.bsvrz.sys.funclib.bitctrl.modell.verkehr.onlinedaten.OdBaustellenSimulation;
import de.bsvrz.sys.funclib.bitctrl.modell.verkehr.onlinedaten.OdBaustellenSimulation.Daten.StauEintrag;
import de.bsvrz.sys.funclib.bitctrl.util.UndefiniertException;
import de.bsvrz.sys.funclib.debug.Debug;

/**
 * berechnet die für eine Baustelle zu erwartetenden Staus.
 *
 * @author BitCtrl Systems GmbH, Peuker
 */
public class BssBaustellenSimulator extends BaustellenSimulator {

	/**
	 * Logger für Debugausgaben.
	 */
	private static final Debug LOGGER = Debug.getLogger();

	/** der Name des Auftraggebers der Simulation. */
	private final String auftragGeber;

	/** eine Bemerkung zum Simulationsauftrag. */
	private final String bemerkung;

	/** optionale vorgegebener Mq vor der Baustelle. */
	private final MessQuerschnittAllgemein berechnungsMq;

	/**
	 * erzeugt eine Instanz der Klasse zur Berechnung von Staus in einer
	 * Baustelle.
	 *
	 * @param auftrag
	 *            der auszuführende Simulationsauftrag
	 * @param parameter
	 *            die Parameter für die Berechnung aus der Baustellenverwaltung
	 */
	public BssBaustellenSimulator(final SimulationsAuftrag auftrag,
			final IBssParameterObserver parameter) {
		super(auftrag, parameter);
		
		LOGGER.finer("Erzeuge BaustellenSimulator für Baustelle: " + auftrag.getBaustelle().getName());
		this.auftragGeber = auftrag.getAuftragGeber();
		this.bemerkung = auftrag.getBemerkung();
		this.berechnungsMq = auftrag.getBerechnungsMq();
	}

	/**
	 * überträgt die Liste der berechneten Staus in einen
	 * Datenverteiler-Datensatz und stellt damit das Ergebnis zur
	 * Veröffentlichung zur Verfügung.
	 *
	 * @param verbindung
	 *            die verwendete Datenverteilerverbindung
	 * @param datenSatz
	 *            der Datensatz, in dem die Ergebnisse abgelegt werden sollen.
	 * @param manuell
	 *            Simulation manuell ausführen
	 */
	public final void ermittleDaten(final ClientDavInterface verbindung, final OdBaustellenSimulation.Daten datenSatz,
			final boolean manuell) {

		boolean simulationErfolgreich = false;
		String fehlerMeldung = null;

		if (getBaustelle().pruefeSimulierbarkeit(manuell)) {

			try {
				final BerechnungsZeitRaumSchnittstelle zeitRaum = new BerechnungsZeitRaum(System.currentTimeMillis(),
						getBaustelle(), getParameter(), manuell);

				if (zeitRaum.isValid()) {
					if (zeitRaum.prognoseMoeglich()) {
						final ZufliessenderVerkehrSchnittstelle zufVerkehr = new ZufliessenderVerkehr(verbindung,
								getBaustelle(), berechnungsMq, zeitRaum);
						prognoseErstellen(zeitRaum, zufVerkehr);
					} else {
						LOGGER.info("Prognosestart- und -endzeit identisch");
						fehlerMeldung = "Prognosestart- und -endzeit identisch";
					}
					simulationErfolgreich = true;
					LOGGER.finer("Liste der Stauobjekte für " + getBaustelle().getName()
					+ " ermittelt - Anzahl: " + getStauListe().size());
				} else {
					fehlerMeldung = MessageFormat.format(BssMessage.WarningUngueltigerSimulationsZeitraum.name(),
							getBaustelle(), zeitRaum);
					LogTools.log(LOGGER, new BetriebsmeldungDaten(getBaustelle().getObjekt()),
							BssMessage.WarningUngueltigerSimulationsZeitraum, getBaustelle(), zeitRaum);
				}
			} catch (final UndefiniertException e) {
				fehlerMeldung = MessageFormat.format(BssMessage.WarningFehlendeSimulationsDaten.name(), getBaustelle(),
						e.getMessage());
				LogTools.log(LOGGER, new BetriebsmeldungDaten(getBaustelle().getObjekt()),
						BssMessage.WarningFehlendeSimulationsDaten, getBaustelle(), e.getMessage());
			}
		}

		datenSatz.setSimulationErfolgreich(simulationErfolgreich);
		datenSatz.setAuftragGeber(auftragGeber);
		datenSatz.setBemerkung(bemerkung);
		datenSatz.setFehlerMeldung(fehlerMeldung);
		for (final BssStau stau : getStauListe()) {
			final StauEintrag eintrag = new OdBaustellenSimulation.Daten.StauEintrag();
			eintrag.setStartZeit(stau.getStartZeit());
			eintrag.setDauer(stau.getDauer());
			eintrag.setMaxLaenge(stau.getMaxLaenge());
			eintrag.setMaxLaengeZeit(stau.getMaxLaengeZeit());
			eintrag.setVerlustZeit(stau.getVerlustZeit());
			datenSatz.addSchritte(eintrag);
		}
	}
}
