/*
 * Segment 6 Intelligente Bewertungsverfahren, SWE 6.2 Baustellensimulation
 * Copyright (C) 2007 BitCtrl Systems GmbH
 *
 * This program is free software; you can redistribute it and/or modify it under
 * the terms of the GNU General Public License as published by the Free Software
 * Foundation; either version 2 of the License, or (at your option) any later
 * version.
 *
 * This program is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
 * FOR A PARTICULAR PURPOSE. See the GNU General Public License for more
 * details.
 *
 * You should have received a copy of the GNU General Public License along with
 * this program; if not, write to the Free Software Foundation, Inc., 51
 * Franklin Street, Fifth Floor, Boston, MA 02110-1301, USA.
 *
 * Contact Information:
 * BitCtrl Systems GmbH
 * Weißenfelser Straße 67
 * 04229 Leipzig
 * Phone: +49 341-490670
 * mailto: info@bitctrl.de
 */

package de.bsvrz.ibv.simbaustelle.simbaustelle;

import de.bsvrz.dav.daf.main.ClientDavInterface;
import de.bsvrz.dav.daf.main.config.ConfigurationObject;
import de.bsvrz.ibv.simbaustelle.parameter.BssParameterObserver;
import de.bsvrz.ibv.simbaustelle.parameter.IBssParameterObserver;
import de.bsvrz.ibv.simbaustelle.verwaltung.BaustellenVerwaltung;
import de.bsvrz.ibv.simbaustelle.verwaltung.SimulationsAuftrag;
import de.bsvrz.sys.funclib.application.StandardApplication;
import de.bsvrz.sys.funclib.application.StandardApplicationRunner;
import de.bsvrz.sys.funclib.bitctrl.modell.ObjektFactory;
import de.bsvrz.sys.funclib.bitctrl.modell.verkehr.objekte.VerkehrModellNetz;
import de.bsvrz.sys.funclib.commandLineArgs.ArgumentList;
import de.bsvrz.sys.funclib.commandLineArgs.ArgumentList.Argument;
import de.bsvrz.sys.funclib.debug.Debug;
import de.bsvrz.sys.funclib.operatingMessage.MessageSender;

/**
 * Die Hauptklasse der Baustellensimulation.
 *
 * @author BitCtrl Systems GmbH, Peuker
 */
public final class BaustellenSimulation implements StandardApplication {

	private Debug logger;

	/** die PID des VerkehrsmodellNetz für das eine Simulation erfolgen soll. */
	private String bstErgTypPid = "ergTyp.baustelle";

	/** die Standardanzahl für Fahrstreifen, wenn keiner ermittelt werden konnte. */
	private int defaultFahrstreifenAnzahl = 2;

	/** die Standardanzahl für Fahrstreifen, wenn keiner ermittelt werden konnte. */
	private boolean polyLinie;

	/**
	 * liefert die PID zur Bestimmung des Baustellenereignistyps.
	 *
	 * @return die PID
	 */
	public static String getBstErgTypPid() {
		return INSTANCE.bstErgTypPid;
	}

	/**
	 * liefert die Standard-Fahrstreofenanzahl.
	 *
	 * @return die Anzahl
	 */
	public static int getDefaultFahrstreifenAnzahl() {
		return INSTANCE.defaultFahrstreifenAnzahl;
	}

	/**
	 * liefert die globale Instanz der Baustellensimulation.
	 *
	 * @return die Instanz
	 */
	public static BaustellenSimulation getInstance() {
		return BaustellenSimulation.INSTANCE;
	}

	/**
	 * liefert Auskunft, ob die per Ganglinie erhaltenen Ergebnisse zwangsweise als
	 * Polylinie verwendet werden.
	 *
	 * @return <code>true</code>, wenn mittels PolyLinie berechnet werden soll
	 */
	public static boolean isPolyLinie() {
		return INSTANCE.polyLinie;
	}

	/**
	 * die Hauptfunktion der Softwareeinheit.
	 *
	 * @param args
	 *            die Kommadozeilenparameter.
	 */
	public static void main(final String[] args) {
		ObjektFactory.getInstanz().registerStandardFactories();
		StandardApplicationRunner.run(BaustellenSimulation.getInstance(), args);
	}

	/** die PID des VerkehrsmodellNetz für das eine Simulation erfolgen soll. */
	private String netzPid;

	/**
	 * die PID des Verwaltungsmoduls mit dem die MDM-Baustellen verwaltet werden.
	 */
	private String mdmVerwaltungPid;

	/**
	 * die Parameterverwaltung zur Bereitstellung der grundlegenden Parameter der
	 * Baustellensimulation des Netzes.
	 */
	private IBssParameterObserver parameter;

	/**
	 * die verwendete Datenverteilerverbindung.
	 */
	private ClientDavInterface verbindung;
 
	/** Lücken in einer Ganglinie sind erlaubt. */
	private boolean allowGaps;

	/**
	 * MQ für die Zuflussberechnung streng nach Anwenderanforderungen suchen.
	 * (Standard ist die ISIS-Methode).
	 */
	private boolean mqAfoStrict;

	/** Bildung des Zeitbereichs für die Simulation streng nach Afo? */
	private boolean zeitbereichAfoStrict;

	/** der maximale Zeitbereich für manuelle Simulationen in Millisekunden. */
	private long maxHorizontManuellMsecs;

	private BaustellenVerwaltung verwaltung;

	/** globale Instanz der Applikation. */
	private static final BaustellenSimulation INSTANCE = new BaustellenSimulation();

	/** privater Konstruktor. */
	private BaustellenSimulation() {
		// Konstruktor wird nur intern aufgerufen
	}

	/**
	 * liefert das Untermodul zr Bereitstellung der Parameter für die Simulation.
	 *
	 * @return des Modul
	 */
	public IBssParameterObserver getParameter() {
		return parameter;
	}

	/**
	 * liefert die verwendete Datenverteilerverbindung.
	 *
	 * @return die Verbindung
	 */
	public ClientDavInterface getVerbindung() {
		return verbindung;
	}

	/**
	 * Die Funktion startet die Untermdule zur Verwaltung der Simulationsparameter
	 * und das Verwaltungsmodul.
	 *
	 * @see de.bsvrz.sys.funclib.application.StandardApplication#initialize(de.bsvrz.dav.daf.main.ClientDavInterface)
	 */
	@Override
	public void initialize(final ClientDavInterface connection) throws Exception {

		MessageSender.getInstance()
				.setApplicationLabel(getClass().getSimpleName() + "_" + connection.getLocalApplicationObject().getId());

		logger.finer("Starte Verwaltungsmodul");

		this.verbindung = connection;
		ObjektFactory.getInstanz().setVerbindung(verbindung);

		ConfigurationObject netzObj = null;
		VerkehrModellNetz netz = null;
		if (netzPid != null) {
			netzObj = (ConfigurationObject) connection.getDataModel().getObject(netzPid);
			if (netzObj == null) {
				throw new IllegalArgumentException(
						"Die PID: \"" + netzPid + "\" für das VerkehrModellNetz beschreibt kein gültiges Objekt");
			}
			if (!netzObj.isOfType("typ.verkehrsModellNetz")) {
				throw new IllegalArgumentException("Der Typ \"" + netzObj.getType() + "\" des Netzobjekt \"" + netzObj
						+ "\" entspricht nicht dem erwarteten \"typ.verkehrsModellNetz\"");
			}
			parameter = new BssParameterObserver(connection, netzObj);
			netz = (VerkehrModellNetz) ObjektFactory.getInstanz().getModellobjekt(netzObj);
		}

		ConfigurationObject mdmVerwaltungObj = null;
		if (mdmVerwaltungPid != null) {
			mdmVerwaltungObj = (ConfigurationObject) connection.getDataModel().getObject(mdmVerwaltungPid);
			if (mdmVerwaltungObj == null) {
				throw new IllegalArgumentException("Die PID: \"" + mdmVerwaltungPid
						+ "\" für Verwaltungsobjekt der MDM-Baustellen beschreibt kein gültiges Objekt");
			}
			if (!mdmVerwaltungObj.isOfType("typ.mdmBaustellenVerwaltung")) {
				throw new IllegalArgumentException("Der Typ \"" + mdmVerwaltungObj.getType()
						+ "\" des MDM-Verwaltungsobjekts \"" + mdmVerwaltungObj
						+ "\" entspricht nicht dem erwarteten \"typ.mdmBaustellenVerwaltung\"");
			}
		}

		if (parameter == null) {
			parameter = new DefaultPrognoseParameter();
		}

		verwaltung = new BaustellenVerwaltung(this, netz, mdmVerwaltungObj);
		verwaltung.init();
	}

	/**
	 * ermittelt, ob Lücken in einer Ganglinie zulässig sind.
	 *
	 * @return den Status
	 */
	public boolean isAllowGaps() {
		return allowGaps;
	}

	/**
	 * ermittelt, ob der MQ für den Zufluss einer Baustelle streng nach
	 * Anwenderforderungen oder nach ISIS-Methode gesucht werden soll.
	 *
	 * @return der Zustand
	 */
	public boolean isMqAfoStrict() {
		return mqAfoStrict;
	}

	/**
	 * Auswertung der Kommandozeilen-Parameter.<br>
	 * Mögliche Parameter sind:
	 * 
	 * <ul>
	 * <li><strong>-netz=&lt;PID&gt;:</strong> PID des verwendeten
	 * VerkehrModellNetz</li>
	 * <li><strong>-MdmBaustellenVerwaltung=&lt;PID&gt;:</strong> PID eines Objekts
	 * zur Verwaltung von MDM-Baustellen</li>
	 * <li><strong>-polyLinie=nein</strong>: bestimmt ob für die verwendeten
	 * Ganglinien Polylinien genutzt werden</li>
	 * <li><strong>-bstErgTyp=ergTyp.baustelle:</strong> das Kommandozeilenargument
	 * zur Definition der PID für die Ermittlung des Baustellen-Ereignistyps</li>
	 * <li><strong>-fahrstreifen=2:</strong> das Kommandozeilenargument zur
	 * Festlegung der Fahrstreifenanzahl, wenn keine definiert ist</li>
	 * <li><strong>-allowGaps=nein:</strong> Lücken einer Ganglinie sind
	 * zulässig</li>
	 * <li><strong>-mqAfoStrict=nein:</strong> Zufluss-MQ-Suche im Afo-strengen
	 * Modus</li>
	 * <li><strong>-zeitBereichStrict=nein:</strong> Zeitbereichsbildung streng nach
	 * Afo</li>
	 * <li><strong>-maxHorizontManuell=0:</strong> das Kommandozeilenargument
	 * beschränkt den maximalen Zeitraum für manuelle Simulationen auf die Anzahl
	 * der festgelegten Tage. Der Standardwert ist 0, was als "keine Beschränkung"
	 * interpretiert wird</li>
	 * </ul>
	 * 
	 * @see de.bsvrz.sys.funclib.application.StandardApplication#parseArguments(de.bsvrz.sys.funclib.commandLineArgs.ArgumentList)
	 */

	@Override
	public void parseArguments(final ArgumentList argumentList) throws Exception {

		logger = Debug.getLogger();

		if (argumentList.hasArgument("-netz")) {
			netzPid = argumentList.fetchArgument("-netz").asNonEmptyString();
		}

		if (argumentList.hasArgument("-MdmBaustellenVerwaltung")) {
			mdmVerwaltungPid = argumentList.fetchArgument("-MdmBaustellenVerwaltung").asNonEmptyString();
		}

		bstErgTypPid = argumentList.fetchArgument("-bstErgTyp=ergTyp.baustelle").asNonEmptyString();
		defaultFahrstreifenAnzahl = argumentList.fetchArgument("-fahrstreifen=2").intValue();
		polyLinie = argumentList.fetchArgument("-polyLinie=nein").booleanValue();
		allowGaps = argumentList.fetchArgument("-allowGaps=nein").booleanValue();
		mqAfoStrict = argumentList.fetchArgument("-mqAfoStrict=nein").booleanValue();
		zeitbereichAfoStrict = argumentList.fetchArgument("-zeitBereichStrict=nein").booleanValue();
		maxHorizontManuellMsecs = argumentList.fetchArgument("-maxHorizontManuell=0").longValue() * 24L * 3600L * 1000L;

		if ((netzPid == null) && (mdmVerwaltungPid == null)) {
			System.out.println("Es wurde weder eine Netz-PID noch eine MDM-Baustellen-Verwaltungs-PID angegeben!");
			System.out.println("Programm wird beendet");
			System.exit(-1);
		}

		if (argumentList.getUnusedArgumentCount() > 0) {
			System.out.println("Unerwartete Argumente:");
			Argument argument;
			while ((argument = argumentList.fetchNextArgument()) != null) {
				System.out.println(argument);
			}
			System.out.println("Programm wird beendet");
			System.exit(-1);
		}
	}

	/**
	 * bestimmt, ob die strikte Zeitbereichsdefinition der Afo verwendet wird.
	 *
	 * @return den Zustand
	 */
	public boolean isZeitbereichAfoStrict() {
		return zeitbereichAfoStrict;
	}

	/**
	 * liefert den maximalen Zeitbereich für manuelle Simulationen in Millisekunden.
	 * Ein Wert &lt;= 0 bedeutet, dass keine Beschränkung wirksam ist.
	 *
	 * @return den Wert
	 */
	public long getMaxHorizontManuellMsecs() {
		return maxHorizontManuellMsecs;
	}

	/**
	 * fügt der Baustellensimulation einen neuen Auftrag hinzu.
	 * 
	 * @param auftrag
	 *            der neue Auftrag
	 */
	public void addAuftrag(SimulationsAuftrag auftrag) {
		verwaltung.addAuftrag(auftrag);
	}
}
