/*
 * Segment 6 Intelligente Bewertungsverfahren, SWE 6.2 Baustellensimulation
 * Copyright (C) 2007 BitCtrl Systems GmbH
 *
 * This program is free software; you can redistribute it and/or modify it under
 * the terms of the GNU General Public License as published by the Free Software
 * Foundation; either version 2 of the License, or (at your option) any later
 * version.
 *
 * This program is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
 * FOR A PARTICULAR PURPOSE. See the GNU General Public License for more
 * details.
 *
 * You should have received a copy of the GNU General Public License along with
 * this program; if not, write to the Free Software Foundation, Inc., 51
 * Franklin Street, Fifth Floor, Boston, MA 02110-1301, USA.
 *
 * Contact Information:
 * BitCtrl Systems GmbH
 * Weißenfelser Straße 67
 * 04229 Leipzig
 * Phone: +49 341-490670
 * mailto: info@bitctrl.de
 */

package de.bsvrz.ibv.simbaustelle.verwaltung;

import java.util.ArrayList;
import java.util.Collections;
import java.util.List;
import java.util.Set;

import de.bsvrz.dav.daf.main.config.SystemObject;
import de.bsvrz.ibv.simbaustelle.prognose.bss.BssBaustellenSimulator;
import de.bsvrz.ibv.simbaustelle.simbaustelle.BaustellenSimulation;
import de.bsvrz.ibv.simbaustelle.simbaustelle.BssMessage;
import de.bsvrz.sys.funclib.bitctrl.daf.BetriebsmeldungDaten;
import de.bsvrz.sys.funclib.bitctrl.daf.LogTools;
import de.bsvrz.sys.funclib.bitctrl.modell.AnmeldeException;
import de.bsvrz.sys.funclib.bitctrl.modell.DatensendeException;
import de.bsvrz.sys.funclib.bitctrl.modell.verkehr.objekte.AeusseresStrassenSegment;
import de.bsvrz.sys.funclib.bitctrl.modell.verkehr.objekte.Baustelle;
import de.bsvrz.sys.funclib.bitctrl.modell.verkehr.objekte.InneresStrassenSegment;
import de.bsvrz.sys.funclib.bitctrl.modell.verkehr.objekte.MessQuerschnittAllgemein;
import de.bsvrz.sys.funclib.bitctrl.modell.verkehr.objekte.StrassenKnoten;
import de.bsvrz.sys.funclib.bitctrl.modell.verkehr.objekte.StrassenSegment;
import de.bsvrz.sys.funclib.bitctrl.modell.verkehr.objekte.StrassenTeilSegment;
import de.bsvrz.sys.funclib.bitctrl.modell.verkehr.onlinedaten.OdBaustellenSimulation;
import de.bsvrz.sys.funclib.bitctrl.modell.verkehr.zustaende.BaustellenStatus;
import de.bsvrz.sys.funclib.bitctrl.util.UndefiniertException;
import de.bsvrz.sys.funclib.debug.Debug;

/**
 * Die Repräsentation einer Baustelle innerhalb der Baustellensimulation.
 *
 * @author BitCtrl Systems GmbH, Peuker
 */
public class BssBaustelle extends Baustelle implements BaustellenSchnittstelle {

	private static final Debug LOGGER = Debug.getLogger();

	/** die beschreibenden Daten der Baustelle. */
	private final BaustellenDaten daten;

	/**
	 * das Datenverteiler-Systemobjekt, das die Baustelle in der Konfiguration
	 * repräsentiert.
	 */
	private final SystemObject bstObjekt;

	/**
	 * der Zeitpunkt der letzzten Berechnung der Baustelle.
	 */
	private long letzteBerechnung;

	/**
	 * die Liste der Segmente des Netzes, dessen Baustellen verwaltet werden sollen.
	 */
	private final Set<StrassenSegment> segmentListe;

	/** der Beobachter für manuelle Simulation. */
	private final BaustellenSimulationsObserver observer;

	/**
	 * erzeugt eine Instanz eines Baustellenobjekts.
	 *
	 * @param segmentListe
	 *            die Liste der Segmente des simulierten Netzes
	 * @param bst
	 *            das Systemobjekt, das die Baustelle in der
	 *            Datenverteiler-Konfiguration repräsentiert
	 */
	public BssBaustelle(final Set<StrassenSegment> segmentListe,
			final SystemObject bst) {
		super(bst);
		this.segmentListe = segmentListe;

		LOGGER.finer("Erzeuge neues Baustellenobjekt: " + bst.getNameOrPidOrId());
		this.bstObjekt = bst;

		getOnlineDatensatz(OdBaustellenSimulation.class)
				.setQuelle(OdBaustellenSimulation.Aspekte.PrognoseNormal.getAspekt(), true);
		try {
			getOnlineDatensatz(OdBaustellenSimulation.class)
					.anmeldenSender(OdBaustellenSimulation.Aspekte.PrognoseNormal.getAspekt());
		} catch (final AnmeldeException e) {
			throw new IllegalStateException(
					"Die Anmeldung als Quelle für die Simulationsdaten der Baustelle " + this + " ist fehlgeschlagen",
					e);
		}

		daten = new BaustellenDaten(this, BaustellenSimulation.getInstance().getParameter());
		observer = new BaustellenSimulationsObserver(this);
	}

	/**
	 * aktualisiert den Zeitpunkt der zyklischen Berechnung der Baustelle.
	 */
	public final void aktualisiereLetzteZyklischeBerechnung() {
		letzteBerechnung = System.currentTimeMillis();
	}

	@Override
	public final int anzahlFahrStreifen(final long position) {

		int result = 0;

		try {
			final List<StrassenSegment> segmente = daten.getSegmente();
			StrassenSegment aktuellesSegment = null;
			double aktuelleLange = 0;
			StrassenTeilSegment aktuellesTeilSegment = null;

			if (segmente.size() > 0) {
				aktuellesSegment = segmente.get(0);
				while (aktuellesSegment != null) {
					double segmentLaenge = 0;
					final List<StrassenTeilSegment> stsListe = new ArrayList<>(
							aktuellesSegment.getStrassenTeilSegmente());
					Collections.reverse(stsListe);
					for (final StrassenTeilSegment sts : stsListe) {
						if (aktuellesTeilSegment == null) {
							segmentLaenge += sts.getLaenge();
							if (segmentLaenge >= daten.getEndOffset()) {
								aktuellesTeilSegment = sts;
								aktuelleLange = ((double) aktuellesSegment.getLaenge()
										- aktuellesTeilSegment.getLaenge()) + daten.getEndOffset();
							}
						} else {
							if (aktuelleLange >= position) {
								aktuellesSegment = null;
								break;
							}
						}
					}

					if (aktuellesSegment instanceof AeusseresStrassenSegment) {
						StrassenSegment naechtesSegment = null;
						final StrassenKnoten knoten = ((AeusseresStrassenSegment) aktuellesSegment).getVonKnoten();
						for (final InneresStrassenSegment iss : knoten.getInnereSegmente()) {
							if (aktuellesSegment.equals(iss.getNachSegment())) {
								if ((iss.getVonSegment() != null) && (iss.getVonSegment().getStrasse() != null)) {
									if (iss.getVonSegment().getStrasse().equals(aktuellesSegment.getStrasse())) {
										naechtesSegment = iss;
										break;
									}
								}
							}
						}
						aktuellesSegment = naechtesSegment;

					} else if (aktuellesSegment instanceof InneresStrassenSegment) {
						aktuellesSegment = ((InneresStrassenSegment) aktuellesSegment).getVonSegment();
					} else {
						aktuellesSegment = null;
					}
				}
			}

			if (aktuellesTeilSegment == null) {
				LOGGER.info("An der Stauposition " + position + " m vor der Baustelle " + this
						+ " konnte kein Teilsegment zur Bestimmung der Fahrstreifenanzahl ermittelt werden.");
			} else {
				result = aktuellesTeilSegment.getAnzahlFahrstreifen();
			}
		} catch (final UndefiniertException e) {
			LOGGER.info("Die Daten zur Bestimmung der Fahrstreifenzahl an der Stauposition " + position
					+ " m der Baustelle " + this + " sind nicht vollständig.", e);
		}

		if (result <= 0) {
			LOGGER.info("An der Stauposition " + position + " m vor der Baustelle " + this
					+ " konnte keine Fahrstreifenanzahl ermittelt werden.");
			result = BaustellenSimulation.getDefaultFahrstreifenAnzahl();
		}

		return result;
	}

	@Override
	public final BaustellenDatenSchnittstelle getDaten() {
		return daten;
	}

	/**
	 * liefert den Zeitpunkt, zu dem die letzte zyklische Berechnung ausgeführt
	 * wurde.
	 *
	 * @return den Zeitpunkt
	 */
	public final long getLetzteZyklischeBerechnung() {
		return letzteBerechnung;
	}

	/**
	 * liefert den vor der Baustelle liegenden Messquerschnitt.
	 *
	 * @return den Messquerschnitt
	 * @throws UndefiniertException
	 *             es konnte kein Messquerschnitt ermittelt werden
	 */
	@Override
	public final MessQuerschnittAllgemein getMessquerschnittDavor() throws UndefiniertException {
		return daten.getMessquerschnittDavor();
	}

	@Override
	public final String getName() {
		return getSystemObjekt().getNameOrPidOrId();
	}

	@Override
	public SystemObject getObjekt() {
		return bstObjekt;
	}

	/**
	 * liefert den aktuelle parametrierten Status der Baustelle.
	 *
	 * @return den Status
	 */
	public final BaustellenStatus getStatus() {
		return daten.getStatus();
	}

	/**
	 * liefert das Systemobjekt, das die Baustelle innerhalb der
	 * Datenverteiler-Konfiguration repräsentiert.
	 *
	 * @return das Objekt
	 */
	public final SystemObject getSystemObjekt() {
		return bstObjekt;
	}

	/**
	 * ermittelt, ob für die Baustelle prinzipiell eine Simulation ausgeführt werden
	 * kann.<br>
	 * EIne Simulation ist nur dann durchführbar, wenn sie den Status GÜLTIG hat und
	 * wenn mindestens einer der ihre Lage beschreibenden Strassensegmente innerhalb
	 * des behandelten Netzes liegt.
	 *
	 * @param manuell
	 *            Simulation manuell ausführen
	 * @return <b><i>true</i></b>, wenn die Simulation ausgeführt werden kann
	 */
	@Override
	public final boolean pruefeSimulierbarkeit(final boolean manuell) {
		boolean result = manuell;

		if (!result) {
			final BaustellenStatus status = daten.getStatus();
			switch (status) {
			case ENTWORFEN:
			case GEPLANT:
			case STORNIERT:
				break;
			case GUELTIG:
				result = true;
				break;
			default:
				break;
			}
		}

		if (result) {
			try {
				result = false;
				for (final StrassenSegment segment : daten.getSegmente()) {
					if (segmentListe.contains(segment)) {
						result = true;
						break;
					}
				}
			} catch (final UndefiniertException e) {
				LOGGER.finer(e.getMessage());
				// die Baustelle ist nicht simulierbar, weil sie keine
				// Straßensegment enthält
			}
		}

		return result;
	}

	/**
	 * ermittelt die Liste der Stauobjekte, die bei entsprechend den
	 * Simulationsvorgaben entstehen und versendet diese an den Datenverteiler.
	 *
	 * @param auftrag
	 *            der auszuführende Auftrag
	 */
	@Override
	public final synchronized void simuliere(final SimulationsAuftrag auftrag) {

		BaustellenSimulation baustellenSimulation = BaustellenSimulation.getInstance();
		final OdBaustellenSimulation.Daten datenSatz = getOnlineDatensatz(OdBaustellenSimulation.class).erzeugeDatum();

		new BssBaustellenSimulator(auftrag, baustellenSimulation.getParameter()).ermittleDaten(baustellenSimulation.getVerbindung(), datenSatz,
				auftrag.isManuell());
		try {
			getOnlineDatensatz(OdBaustellenSimulation.class)
					.sendeDaten(OdBaustellenSimulation.Aspekte.PrognoseNormal.getAspekt(), datenSatz);
		} catch (final DatensendeException e) {
			LogTools.log(LOGGER, new BetriebsmeldungDaten(bstObjekt), BssMessage.ErrorSendeSimulationsErgebnis,
					bstObjekt, e.getMessage());
		}
	}

	/** entfernt die Datenanmeldungen der Baustellen-Instanz. */
	public void disconnect() {
		observer.disconnect();
		getOnlineDatensatz(OdBaustellenSimulation.class)
				.abmeldenSender(OdBaustellenSimulation.Aspekte.PrognoseNormal.getAspekt());
	}
	
	@Override
	public String toString() {
		return bstObjekt.toString();
	}
}
