/*
 * Segment 6 Intelligente Bewertungsverfahren, SWE 6.2 Baustellensimulation
 * Copyright (C) 2007 BitCtrl Systems GmbH
 *
 * This program is free software; you can redistribute it and/or modify it under
 * the terms of the GNU General Public License as published by the Free Software
 * Foundation; either version 2 of the License, or (at your option) any later
 * version.
 *
 * This program is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
 * FOR A PARTICULAR PURPOSE. See the GNU General Public License for more
 * details.
 *
 * You should have received a copy of the GNU General Public License along with
 * this program; if not, write to the Free Software Foundation, Inc., 51
 * Franklin Street, Fifth Floor, Boston, MA 02110-1301, USA.
 *
 * Contact Information:
 * BitCtrl Systems GmbH
 * Weißenfelser Straße 67
 * 04229 Leipzig
 * Phone: +49 341-490670
 * mailto: info@bitctrl.de
 */

package de.bsvrz.ibv.simbaustelle.verwaltung;

import java.util.ArrayList;
import java.util.Collections;
import java.util.List;

import de.bsvrz.dav.daf.main.ClientDavInterface;
import de.bsvrz.dav.daf.main.ClientSenderInterface;
import de.bsvrz.dav.daf.main.Data;
import de.bsvrz.dav.daf.main.DataDescription;
import de.bsvrz.dav.daf.main.OneSubscriptionPerSendData;
import de.bsvrz.dav.daf.main.ResultData;
import de.bsvrz.dav.daf.main.SendSubscriptionNotConfirmed;
import de.bsvrz.dav.daf.main.SenderRole;
import de.bsvrz.dav.daf.main.config.Aspect;
import de.bsvrz.dav.daf.main.config.AttributeGroup;
import de.bsvrz.dav.daf.main.config.DataModel;
import de.bsvrz.dav.daf.main.config.SystemObject;
import de.bsvrz.ibv.simbaustelle.prognose.mdm.MdmBaustellenSimulator;
import de.bsvrz.ibv.simbaustelle.simbaustelle.BaustellenSimulation;
import de.bsvrz.ibv.simbaustelle.simbaustelle.BssMessage;
import de.bsvrz.sys.funclib.bitctrl.daf.BetriebsmeldungDaten;
import de.bsvrz.sys.funclib.bitctrl.daf.LogTools;
import de.bsvrz.sys.funclib.bitctrl.modell.verkehr.objekte.AeusseresStrassenSegment;
import de.bsvrz.sys.funclib.bitctrl.modell.verkehr.objekte.InneresStrassenSegment;
import de.bsvrz.sys.funclib.bitctrl.modell.verkehr.objekte.MessQuerschnittAllgemein;
import de.bsvrz.sys.funclib.bitctrl.modell.verkehr.objekte.StrassenKnoten;
import de.bsvrz.sys.funclib.bitctrl.modell.verkehr.objekte.StrassenSegment;
import de.bsvrz.sys.funclib.bitctrl.modell.verkehr.objekte.StrassenTeilSegment;
import de.bsvrz.sys.funclib.bitctrl.util.UndefiniertException;
import de.bsvrz.sys.funclib.debug.Debug;

/**
 * Die Repräsentation einen MDM-Baustelle für die Simulation.
 * 
 * @author BitCtrl Systems GmbH, Uwe Peuker
 */
public class MdmBaustelle implements BaustellenSchnittstelle, ClientSenderInterface {

	private static final Debug LOGGER = Debug.getLogger();
	
	private MdmBaustellenSimulationsObserver observer;
	private final SystemObject bstObjekt;
	private ClientDavInterface connection;
	private DataDescription prognoseDesc;

	private DataDescription prognoseDetailDesc;

	private BaustellenDatenSchnittstelle daten;

	public MdmBaustelle(SystemObject bstObject) {

		LOGGER.finer("Erzeuge neues Baustellenobjekt: " + bstObject.getNameOrPidOrId());

		this.connection = BaustellenSimulation.getInstance().getVerbindung();
		this.bstObjekt = bstObject;

		DataModel model = connection.getDataModel();
		
		AttributeGroup atg = model.getAttributeGroup("atg.baustellenSimulation");
		Aspect asp = model.getAspect("asp.prognoseNormal");
		prognoseDesc = new DataDescription(atg, asp);
		try {
			connection.subscribeSender(this, bstObjekt, prognoseDesc, SenderRole.source());
		} catch (OneSubscriptionPerSendData e) {
			LOGGER.warning(e.getLocalizedMessage());
		}

		atg = model.getAttributeGroup("atg.baustellenSimulationDetailliert");
		prognoseDetailDesc = new DataDescription(atg, asp);
		try {
			connection.subscribeSender(this, bstObjekt, prognoseDetailDesc, SenderRole.source());
		} catch (OneSubscriptionPerSendData e) {
			LOGGER.warning(e.getLocalizedMessage());
		}

		daten = new MdmBaustellenDaten(connection, this);
		observer = new MdmBaustellenSimulationsObserver(this);
	}

	@Override
	public int anzahlFahrStreifen(long position) throws UndefiniertException {
		int result = 0;

		try {
			final List<StrassenSegment> segmente = daten.getSegmente();
			StrassenSegment aktuellesSegment = null;
			double aktuelleLange = 0;
			StrassenTeilSegment aktuellesTeilSegment = null;

			if (segmente.size() > 0) {
				aktuellesSegment = segmente.get(0);
				while (aktuellesSegment != null) {
					double segmentLaenge = 0;
					final List<StrassenTeilSegment> stsListe = new ArrayList<>(
							aktuellesSegment.getStrassenTeilSegmente());
					Collections.reverse(stsListe);
					for (final StrassenTeilSegment sts : stsListe) {
						if (aktuellesTeilSegment == null) {
							segmentLaenge += sts.getLaenge();
							if (segmentLaenge >= daten.getEndOffset()) {
								aktuellesTeilSegment = sts;
								aktuelleLange = ((double) aktuellesSegment.getLaenge()
										- aktuellesTeilSegment.getLaenge()) + daten.getEndOffset();
							}
						} else {
							if (aktuelleLange >= position) {
								aktuellesSegment = null;
								break;
							}
						}
					}

					if (aktuellesSegment instanceof AeusseresStrassenSegment) {
						StrassenSegment naechtesSegment = null;
						final StrassenKnoten knoten = ((AeusseresStrassenSegment) aktuellesSegment).getVonKnoten();
						for (final InneresStrassenSegment iss : knoten.getInnereSegmente()) {
							if (aktuellesSegment.equals(iss.getNachSegment())) {
								if ((iss.getVonSegment() != null) && (iss.getVonSegment().getStrasse() != null)) {
									if (iss.getVonSegment().getStrasse().equals(aktuellesSegment.getStrasse())) {
										naechtesSegment = iss;
										break;
									}
								}
							}
						}
						aktuellesSegment = naechtesSegment;

					} else if (aktuellesSegment instanceof InneresStrassenSegment) {
						aktuellesSegment = ((InneresStrassenSegment) aktuellesSegment).getVonSegment();
					} else {
						aktuellesSegment = null;
					}
				}
			}

			if (aktuellesTeilSegment == null) {
				LOGGER.info("An der Stauposition " + position + " m vor der Baustelle " + this
						+ " konnte kein Teilsegment zur Bestimmung der Fahrstreifenanzahl ermittelt werden.");
			} else {
				result = aktuellesTeilSegment.getAnzahlFahrstreifen();
			}
		} catch (final UndefiniertException e) {
			LOGGER.info("Die Daten zur Bestimmung der Fahrstreifenzahl an der Stauposition " + position
					+ " m der Baustelle " + this + " sind nicht vollständig.", e);
		}

		if (result <= 0) {
			LOGGER.info("An der Stauposition " + position + " m vor der Baustelle " + this
					+ " konnte keine Fahrstreifenanzahl ermittelt werden.");
			result = BaustellenSimulation.getDefaultFahrstreifenAnzahl();
		}

		return result;
	}

	@Override
	public BaustellenDatenSchnittstelle getDaten() {
		return daten;
	}

	@Override
	public MessQuerschnittAllgemein getMessquerschnittDavor() throws UndefiniertException {
		throw new UndefiniertException("Der Messquerschnitt für den zufließenden Verkehr muss bei Mdm-Baustellen im Simulationsauftrag übergeben werden!");
	}

	@Override
	public String getName() {
		return bstObjekt.getNameOrPidOrId();
	}

	@Override
	public SystemObject getObjekt() {
		return bstObjekt;
	}

	@Override
	public boolean pruefeSimulierbarkeit(boolean manuell) {
		return true;
	}

	@Override
	public void simuliere(SimulationsAuftrag auftrag) {
		
		BaustellenSimulation baustellenSimulation = BaustellenSimulation.getInstance();
		final Data datenSatz = connection.createData(prognoseDesc.getAttributeGroup());
		final Data datenSatzDetail = connection.createData(prognoseDetailDesc.getAttributeGroup());

		new MdmBaustellenSimulator(auftrag, baustellenSimulation.getParameter()).ermittleDaten(connection, datenSatz,
				datenSatzDetail);
		try {
			connection.sendData(new ResultData(bstObjekt, prognoseDesc, connection.getTime(), datenSatz));
		} catch (final SendSubscriptionNotConfirmed e) {
			LogTools.log(LOGGER, new BetriebsmeldungDaten(bstObjekt), BssMessage.ErrorSendeSimulationsErgebnis,
					bstObjekt, e.getMessage());
		}

		try {
			connection.sendData(new ResultData(bstObjekt, prognoseDetailDesc, connection.getTime(), datenSatzDetail));
		} catch (final SendSubscriptionNotConfirmed e) {
			LogTools.log(LOGGER, new BetriebsmeldungDaten(bstObjekt), BssMessage.ErrorSendeSimulationsErgebnis,
					bstObjekt, e.getMessage());
		}
	}

	public void disconnect() {
		daten.disconnect();
		observer.disconnect();
		connection.unsubscribeSender(this, bstObjekt, prognoseDesc);
		connection.unsubscribeSender(this, bstObjekt, prognoseDetailDesc);
	}

	@Override
	public void dataRequest(SystemObject object, DataDescription dataDescription, byte state) {
		// wird nicht betrachtet
	}

	@Override
	public boolean isRequestSupported(SystemObject object, DataDescription dataDescription) {
		return false;
	}
	
	@Override
	public String toString() {
		return bstObjekt.toString();
	}
}
