/*
 * Segment 6 Intelligente Bewertungsverfahren, SWE 6.2 Baustellensimulation
 * Copyright (C) 2007 BitCtrl Systems GmbH
 *
 * This program is free software; you can redistribute it and/or modify it under
 * the terms of the GNU General Public License as published by the Free Software
 * Foundation; either version 2 of the License, or (at your option) any later
 * version.
 *
 * This program is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
 * FOR A PARTICULAR PURPOSE. See the GNU General Public License for more
 * details.
 *
 * You should have received a copy of the GNU General Public License along with
 * this program; if not, write to the Free Software Foundation, Inc., 51
 * Franklin Street, Fifth Floor, Boston, MA 02110-1301, USA.
 *
 * Contact Information:
 * BitCtrl Systems GmbH
 * Weißenfelser Straße 67
 * 04229 Leipzig
 * Phone: +49 341-490670
 * mailto: info@bitctrl.de
 */

package de.bsvrz.ibv.simbaustelle.verwaltung;

import java.util.concurrent.TimeUnit;

import de.bsvrz.dav.daf.main.ClientDavInterface;
import de.bsvrz.dav.daf.main.ClientReceiverInterface;
import de.bsvrz.dav.daf.main.Data;
import de.bsvrz.dav.daf.main.Data.NumberValue;
import de.bsvrz.dav.daf.main.DataDescription;
import de.bsvrz.dav.daf.main.ReceiveOptions;
import de.bsvrz.dav.daf.main.ReceiverRole;
import de.bsvrz.dav.daf.main.ResultData;
import de.bsvrz.dav.daf.main.config.Aspect;
import de.bsvrz.dav.daf.main.config.AttributeGroup;
import de.bsvrz.dav.daf.main.config.DataModel;
import de.bsvrz.dav.daf.main.config.SystemObject;
import de.bsvrz.ibv.simbaustelle.simbaustelle.BaustellenSimulation;
import de.bsvrz.sys.funclib.bitctrl.modell.ObjektFactory;
import de.bsvrz.sys.funclib.bitctrl.modell.verkehr.objekte.MessQuerschnittAllgemein;
import de.bsvrz.sys.funclib.debug.Debug;

/**
 * eine Hilfsobjekt zur Verarbeitung der Kommandos zur Ausführung einer
 * Simulation einen MDM-Baustelle auf Nutzeranforerung.<br>
 * Die Klasse wertet die über die Attributgruppe
 * "atg.baustellenSimulationStartenMitParametern" empfangenen Daten aus und löst
 * eine sofortige Berechnung der entsprechenden Baustelle aus.
 *
 * @author BitCtrl Systems GmbH, Peuker
 */
public class MdmBaustellenSimulationsObserver implements ClientReceiverInterface {

	/** Logger für Debugausgaben. */
	static final Debug LOGGER = Debug.getLogger();

	/** die Attributgruppe zum Empfang des Kommandos zum Start der Simulation. */
	private AttributeGroup startAtg;

	/** der erwartetet Aspekt. */
	private Aspect startAsp;

	/**
	 * die Baustelle, die gesteuert werden soll.
	 */
	private final MdmBaustelle baustelle;

	private ClientDavInterface connection;

	/**
	 * erzeugt eine Instanz des Steuerungsobjekts.
	 *
	 * @param baustelle
	 *            die zu steuernde Baustelle
	 */
	public MdmBaustellenSimulationsObserver(final MdmBaustelle baustelle) {

		this.baustelle = baustelle;
		this.connection = BaustellenSimulation.getInstance().getVerbindung();

		final DataModel model = baustelle.getObjekt().getDataModel();
		startAtg = model.getAttributeGroup("atg.baustellenSimulationStartenMitParametern");
		startAsp = model.getAspect("asp.senden");

		connection.subscribeReceiver(this, baustelle.getObjekt(), new DataDescription(startAtg, startAsp),
				ReceiveOptions.normal(), ReceiverRole.drain());
	}

	/**
	 * Die Funktion empfängt den Datensatz zur Ausführung einer Simulation und führt
	 * die Simulation der Baustelle aus.
	 *
	 * @see de.bsvrz.dav.daf.main.ClientReceiverInterface#update(de.bsvrz.dav.daf.main.ResultData[])
	 */
	@Override
	public final void update(final ResultData[] results) {
		for (final ResultData result : results) {
			LOGGER.finer("Befehl zum manuellen Simulationsstart empfangen", result);

			final Data data = result.getData();
			if (data != null) {
				if (result.getDataDescription().getAttributeGroup().equals(startAtg)) {
					final String name = data.getTextValue("Name").getText();
					final String bemerkung = data.getTextValue("Bemerkung").getText();
					final SystemObject mqObj = data.getReferenceValue("BerechnungsMQ").getSystemObject();
					MessQuerschnittAllgemein berechnungsMq = null;
					if (mqObj != null) {
						berechnungsMq = (MessQuerschnittAllgemein) ObjektFactory.getInstanz().getModellobjekt(mqObj);
					}
					long startZeit = data.getTimeValue("StartZeit").getMillis();
					long endZeit = data.getTimeValue("EndZeit").getMillis();

					Long restKapazitaet = null;
					NumberValue value = data.getScaledValue("RestKapazität");
					if (value.isState()) {
						LOGGER.warning(baustelle + ": Die Restkapazität hat keinen gültigen Wert");
					} else {
						restKapazitaet = value.longValue();
					}

					Long schrittWeiteInMsec = null;
					value = data.getScaledValue("Schrittweite");
					if (value.isState()) {
						LOGGER.warning(baustelle + ": Die Schrittweite hat keinen gültigen Wert");
					} else {
						schrittWeiteInMsec = TimeUnit.MINUTES.toMillis(value.longValue());
					}

					if (berechnungsMq != null && schrittWeiteInMsec != null && restKapazitaet != null) {
						ManuellerAuftrag auftrag = new ManuellerAuftrag(baustelle, name, bemerkung, berechnungsMq);
						auftrag.setStartZeit(startZeit);
						auftrag.setEndZeit(endZeit);
						auftrag.setSchrittWeiteInMsec(schrittWeiteInMsec);
						auftrag.setRestKapazitaet(restKapazitaet);
						BaustellenSimulation.getInstance()
								.addAuftrag(auftrag);
					}
				}
			}
		}
	}

	/** entfernt die Datenanmeldungen des Observers. */
	public void disconnect() {
		connection.unsubscribeReceiver(this, baustelle.getObjekt(), new DataDescription(startAtg, startAsp));
	}
}
